from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.step import Stepper


class CmdCleanup(Cmd):
    def run(self) -> None:
        stepper = Stepper()

        with stepper.step('fix hooks') as step:
            step.run_cmd('chmod 755 -R .git/hooks')

        with stepper.step('docker login') as step:
            step.run_cmd('docker login gitlab.neroelectronics.by:5050 -u unic-lab-developers -p Fqp99SoNxBYPA8ZPaXy2')

        with stepper.step('git sync') as step:
            step.run_cmd('git submodule init')
            step.run_cmd('git submodule update --remote')

        with stepper.step('pipenv sync') as step:
            step.run_cmd('pipenv sync --dev')
            step.run_cmd('pipenv clean')

        with stepper.step('ulpytool install') as step:
            step.run_cmd('ulpytool install')



import argparse
from ul_py_tool.utils.write_stdout import write_stdout

from typing import List


def arg_str2list(value: str) -> List[str]:  # Because we process dict of KeyValue
    if not isinstance(value, str):
        raise argparse.ArgumentTypeError('invalid type')  # because required to be key=value,key=...
    res = []
    kv_list = value.split(",")
    for kv in kv_list:
        if not kv:
            continue  # ignore if something went wrong and we got like a=b,,,,,,
        split_kv = kv.split("=")
        if len(split_kv) != 2:
            write_stdout(f'cant process {kv} kv, skipping...')  # same as above but a====b and etc.
            continue
        if not split_kv[1]:
            write_stdout(f'No value was provided for {split_kv[0]}, skipping...')
            continue
        res.append(kv)
    return res



import os

from ul_py_tool.utils.write_stdout import write_stdout


def copy_if_destination_is_absent(fp_from: str, fp_to: str, replace: bool = False) -> None:
    assert os.path.isfile(fp_from)
    if replace or not os.path.isfile(fp_to):
        os.makedirs(os.path.dirname(fp_to), exist_ok=True)
        with open(fp_from, 'rt') as f_from, open(fp_to, 'wt+') as f_to:
            f_to.writelines(f_from.readlines())
        write_stdout(f'file "{fp_to}" created')



import argparse
import os.path

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.commands.conf import PIPFILE
from ul_py_tool.utils.colors import NC, FG_GRAY, FG_GREEN, FG_RED
from ul_py_tool.utils.pipfile import Pipfile
from ul_py_tool.utils.write_stdout import write_stdout


class CmdOutdated(Cmd):
    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        pass

    def run(self) -> None:
        pipfile = os.path.join(os.getcwd(), PIPFILE)
        ppfile = Pipfile.load(pipfile)

        max_len_pack = 0
        max_len_ver = 7
        for pack in ppfile.all_packages.values():
            max_len_pack = max(max_len_pack, len(pack.name))
            max_len_ver = max(max_len_ver, len(pack.clean_version))

        write_stdout(f'\n        {FG_GRAY}|{NC} {"Name": <{max_len_pack}} {FG_GRAY}|{NC} {"Current": >{max_len_ver}} {FG_GRAY}|{NC} Actual')
        write_stdout(f' {FG_GRAY}{"-" * 7}|{"-" * (1 + max_len_pack + 1)}|{"-" * (1 + max_len_ver + 1)}|{"-" * 16}{NC}')
        ln = len(ppfile.all_packages)
        has_outdated = False
        for i, pack in enumerate(ppfile.all_packages.values()):
            if len(pack.available_versions) > 0:
                has_outdated = True
                write_stdout(f'  {FG_GRAY}{i + 1:0>2}/{ln:0<2}{NC} {FG_GRAY}|{NC}'
                             f' {FG_RED}{pack.name: <{max_len_pack}}{NC} {FG_GRAY}|{NC}'
                             f' {FG_RED}{pack.clean_version: >{max_len_ver}}{NC} {FG_GRAY}|{NC}'
                             f' {FG_GREEN}{pack.available_versions[-1]}{NC}')
            else:
                write_stdout(f'  {FG_GRAY}{i + 1:0>2}/{ln:0<2}{NC} {FG_GRAY}|{NC}'
                             f' {pack.name: <{max_len_pack}} {FG_GRAY}|{NC}'
                             f' {pack.clean_version: >{max_len_ver}} {FG_GRAY}|{NC}'
                             f' {FG_GRAY}{pack.clean_version}{NC}')
        write_stdout('' if has_outdated else f'{FG_GREEN}NOTHING TO UPDATE{NC}')
        exit(int(has_outdated))



import argparse
import os.path
import subprocess
from typing import List

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_files_glob import arg_files_glob, arg_file_glob_compile_files, arg_files_print, arg_file_info
from ul_py_tool.utils.colors import FG_RED, NC
from ul_py_tool.utils.step import Stepper, StepError
from ul_py_tool.utils.write_stdout import write_stdout


class CmdTest(Cmd):
    print_files: int
    py_file_ignore_lists: List[List[str]]
    py_file_lists: List[List[str]]

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        finder = arg_files_glob(ignore_absent=True)
        parser.add_argument('--print-files', dest='print_files', type=int, required=False, default=10)
        parser.add_argument('--files', dest='py_file_lists', help=f'{FG_RED}DEPRECATED!!!{NC}', nargs='+', type=finder, required=False, default=[])
        parser.add_argument('--ignore', dest='py_file_ignore_lists', nargs='+', type=finder, required=False, default=[
            finder('*/**/main.py'),
            finder('.*/**/*.py'),
            finder('*/**/__*__.py'),
        ])

    def _check_pytest_response(self, res: 'subprocess.CompletedProcess[bytes]') -> None:
        if res.returncode == 0:
            return
        if res.returncode == 5:
            return
        res.check_returncode()

    def _get_ini_cwd(self) -> str:
        cwd = os.getcwd()
        while cwd and len(cwd) > 5:
            if os.path.isfile(os.path.join(cwd, 'pytest.ini')):
                return cwd
            cwd = os.path.dirname(cwd)
        return os.getcwd()

    def run(self) -> None:
        stepper = Stepper()
        cwd = self._get_ini_cwd()
        tmp_dir = os.path.join(cwd, ".tmp")

        with stepper.step('PREPARE'):
            py_files, _1 = arg_file_glob_compile_files(self.py_file_lists, self.py_file_ignore_lists)
            if len(py_files) > 0:
                files_to_test = ",\n   ".join(arg_file_info(f) for f in py_files)
                raise StepError(f'--files param is deprecated!! please move your tests to __tests__dirs. files to test: \n   {files_to_test}')

            finder = arg_files_glob(ignore_absent=True)
            py_files, ignored_files = arg_file_glob_compile_files(
                include=[
                    finder('**/__tests__/*.py'),
                    finder('**/__tests__/**/*.py'),
                    finder('__tests__/*.py'),
                    finder('__tests__/**/*.py'),
                ],
                exclude=[
                    *self.py_file_ignore_lists,
                    finder('**/_*.py'),
                    finder('_*.py'),
                ],
            )

        with stepper.step('PYTEST') as stp:
            if len(py_files) == 0:
                write_stdout('no files to test')
                stp.skip()
            else:
                if self.print_files:
                    write_stdout('\n')
                    arg_files_print(self.print_files, py_files, ignored_files=ignored_files, name='testing files')
                    write_stdout('\n')
                    arg_files_print(self.print_files, ignored_files, name='ignored testing files')
                    write_stdout('\n')
                stp.run_cmd(
                    f'pytest -rxXs '
                    f'--cov={cwd} --no-cov-on-fail --cov-report html:{os.path.join(tmp_dir, "coverage")} '
                    f'-o cache_dir={tmp_dir} {" ".join(py_files)}',
                    chk=self._check_pytest_response,
                )


def test_example() -> None:
    assert 1 == 1



import os
import subprocess
import sys
from typing import List, NamedTuple, Tuple, Union

from ul_py_tool.utils.colors import FG_GREEN, NC, FG_RED

PIPE = 'pipe'
STDOUT = 'stdout'
DEVNULL = 'dev_null'

PATH_CONST__CWD = os.path.dirname(os.path.dirname(__file__))


STDERR_DESCRIPTOR_MAP = {
    PIPE: subprocess.PIPE,
    STDOUT: subprocess.STDOUT,
    DEVNULL: subprocess.DEVNULL,
}


class CommandResult(NamedTuple):
    code: int
    stdout: str

    def display(self, disable: bool = False) -> None:
        if disable:
            return

        sys.stdout.write(f'{self}\n')
        sys.stdout.flush()

    def __str__(self) -> str:
        return self.stdout

    def __repr__(self) -> str:
        return self.__str__()


def add_line_pref(some_str: str, pref: str) -> str:
    if not len(pref):
        return some_str

    result_str_lines = list()
    for str_line in some_str.split('\n'):
        if str_line.startswith(pref):
            result_str_lines.append(str_line)
        else:
            result_str_lines.append(f'{pref}{str_line}')

    return '\n'.join(result_str_lines)


def execute_command(command: List[str], cwd: str, silent: bool, stderr: int = subprocess.STDOUT) -> Tuple[int, str]:
    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=stderr, cwd=cwd, universal_newlines=True)  # nosec
    stdout: List[str] = []
    while process.poll() is None:
        line = process.stdout.readline()  # type: ignore
        stdout.append(line)
        if not silent:
            sys.stdout.write(line.strip() + '\r\n')
    result_stdout = ''.join(stdout)
    return process.returncode, result_stdout


def run_command(
    command: Union[List[str], str],
    ignore_errors: bool = False,
    silent: bool = False,
    msg_pref: str = '',
    cwd: str = PATH_CONST__CWD,
    stderr: str = STDOUT,
    simple: bool = False,
) -> CommandResult:
    assert stderr in STDERR_DESCRIPTOR_MAP, f'{stderr} given'
    if isinstance(command, str):
        command = [command]
    if not silent:
        for cmd in command:
            sys.stdout.write(f'{msg_pref}{FG_GREEN}(f"RUN::   {cmd}"){NC}\n')
            sys.stdout.flush()
    if simple:
        try:
            result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=PATH_CONST__CWD, shell=True, check=True)  # nosec
        except subprocess.CalledProcessError as e:
            if not silent:
                print(f'{msg_pref}{add_line_pref(e.stdout.decode("utf-8"), msg_pref)}')  # noqa
                print(f'{msg_pref}FAIL:\n{add_line_pref(e.stderr.decode("utf-8"), msg_pref)}')  # noqa
            if ignore_errors:
                return CommandResult(code=e.returncode, stdout=e.stdout.decode('utf-8'))
            raise

        if not silent:
            print(result.stdout.decode('utf-8'))  # noqa
            print(f'{msg_pref}WARNINIG:\n{add_line_pref(result.stderr.decode("utf-8"), msg_pref)}')  # noqa

        cmd_result = CommandResult(code=result.returncode, stdout=result.stdout.decode('utf-8'))

        return cmd_result

    stderr_descriptor = STDERR_DESCRIPTOR_MAP[stderr]
    return_code, stdout = execute_command(command, cwd=cwd, silent=silent, stderr=stderr_descriptor)
    if return_code != 0:
        if not silent:
            sys.stdout.write(f'{msg_pref}{FG_RED}"FAIL:"){NC}\n{add_line_pref(stdout, msg_pref)}')
            sys.stdout.flush()
        if ignore_errors:
            return CommandResult(code=return_code, stdout=stdout)
        raise Exception(f"command \"{command}\" exited with code {return_code}")

    cmd_result = CommandResult(code=return_code, stdout=stdout)

    return cmd_result



import argparse
import importlib
import logging
import os
import sys
from typing import Mapping, Type, Optional, Union

from pydantic import BaseModel

from ul_py_tool.utils.colors import FG_RED, NC, FG_YELLOW
from ul_py_tool.utils.step import StepError
from ul_py_tool.utils.write_stdout import write_stdout


class Cmd(BaseModel):
    cmd: str
    log: str

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        pass

    def run(self) -> None:
        write_stdout('be brave!')

    @staticmethod
    def main(commands: Mapping[str, Union[Type['Cmd'], str]], *, cwd: Optional[str] = None) -> None:
        """

        :rtype: object
        """
        if cwd is not None:
            assert os.path.abspath(os.path.normpath(cwd)) == os.path.abspath(os.path.normpath(os.getcwd())), f'script must run from cwd "{cwd}". from "{os.getcwd()}" was ran'
            sys.path.append(cwd)

        parser = argparse.ArgumentParser()
        parser.add_argument('--log', type=str, default='INFO', required=False)
        cmd_parser = parser.add_subparsers(required=True, dest='cmd')

        cmd_name = sys.argv[1] if len(sys.argv) > 1 else None
        cmd = commands.get(cmd_name or '', None)
        if cmd is None:
            write_stdout(f'{FG_RED}command "{cmd_name}" is not supported. must be on of ({", ".join(commands.keys())}){NC}')
            exit(1)

        assert cmd_name is not None

        if not isinstance(cmd, str):
            cmd_class = cmd
        else:
            assert ':' in cmd
            cmd = cmd.strip()
            cmd_segm = cmd.split(':')
            assert len(cmd_segm) == 2
            mdl = importlib.import_module(cmd_segm[0].strip())
            cmd_class = getattr(mdl, cmd_segm[1].strip())
        assert cmd_class is not None
        assert issubclass(cmd_class, Cmd)
        cmd_class.add_parser_args(cmd_parser.add_parser(cmd_name))

        args = dict(vars(parser.parse_args()))
        log = args['log'].upper()

        logging.basicConfig(level=getattr(logging, log), format='ulpytool :: %(asctime)s :: %(levelname)s :: %(message)s')

        try:
            cmd_instance = cmd_class(**args)
            cmd_instance.run()
        except KeyboardInterrupt:
            write_stdout(f'\n{FG_YELLOW}...interrupted!{NC}')
            sys.exit(1)
        except StepError as e:
            msg = str(e)
            if msg:
                write_stdout(msg)
            write_stdout(f'{FG_RED}FAILED{NC}')
            sys.exit(1)



import argparse
import os.path
from datetime import datetime
from enum import Enum
from typing import Tuple

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_file_exists import arg_file_exists
from ul_py_tool.utils.colors import FG_GREEN, NC
from ul_py_tool.utils.step import Stepper
from ul_py_tool.utils.write_stdout import write_stdout


class CmdServiceVersionType(Enum):
    PATCH = 'patch'
    MINOR = 'minor'
    MAJOR = 'major'


class CmdServiceVersion(Cmd):
    service_version_file: str
    type: CmdServiceVersionType

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--service-version-file', dest='service_version_file', type=arg_file_exists, default=os.path.join(os.getcwd(), '.service-version'), required=False)

    def _increase_version(self) -> str:
        write_stdout(f'read "{self.service_version_file}"')
        with open(self.service_version_file, 'rt') as f:
            version = f.read()

        version_segm: Tuple[int, int, int] = tuple(int(v) for v in version.split('.'))  # type: ignore

        if self.type is CmdServiceVersionType.PATCH:
            new_version_segm = (version_segm[0], version_segm[1], version_segm[2] + 1)
            write_stdout(f'{version} -> {new_version_segm[0]}.{new_version_segm[1]}.{FG_GREEN}{new_version_segm[2]}{NC}')
        elif self.type is CmdServiceVersionType.MINOR:
            new_version_segm = (version_segm[0], version_segm[1] + 1, 0)
            write_stdout(f'{version} -> {new_version_segm[0]}.{FG_GREEN}{new_version_segm[1]}.{new_version_segm[2]}{NC}')
        elif self.type is CmdServiceVersionType.MAJOR:
            new_version_segm = (version_segm[0] + 1, 0, 0)
            write_stdout(f'{version} -> {FG_GREEN}{new_version_segm[0]}.{new_version_segm[1]}.{new_version_segm[2]}{NC}')
        else:
            raise NotImplementedError(f'type {self.type} was not implemented for increasing version command')
        new_version = ".".join(str(i) for i in new_version_segm)

        with open(self.service_version_file, 'wt') as f:
            f.write(new_version)
            write_stdout(f'Changes saved to {self.service_version_file}')

        return new_version

    def run(self) -> None:
        stepper = Stepper()

        with stepper.step('INCREASE VERSION', print_error=True):
            new_version = self._increase_version()

        files_to_add = [self.service_version_file]

        with stepper.step('GIT') as stp:
            stp.run_cmd(['git', 'reset'])
            stp.run_cmd(['git', 'add', *files_to_add])
            stp.run_cmd(['git', 'commit', '-m', new_version])
            stp.run_cmd(['git', 'tag', new_version])


class CmdServiceVersionMinor(CmdServiceVersion):
    type: CmdServiceVersionType = CmdServiceVersionType.MINOR


class CmdServiceVersionPatch(CmdServiceVersion):
    type: CmdServiceVersionType = CmdServiceVersionType.PATCH


class CmdServiceVersionMajor(CmdServiceVersion):
    type: CmdServiceVersionType = CmdServiceVersionType.MAJOR



from typing import List, Dict, Union, Any, Optional, Tuple, Iterable

from pydantic import BaseModel

MAP_CHOICES_RESTART = {
    "Always": "always",
    "OnFailure": "on-failure",
    "Never": "no",
}
KUBERNETES_OBJECTS__TRANSFORM_TO_DCF__TYPE = ['Deployment', 'StatefulSet', 'Job']
KUBERNETES_OBJECTS__DEPENDS_ON_OBJECT__TYPE = ['StatefulSet']


class DockerComposeService(BaseModel):
    name: str
    kubernetes_type: str
    replicas: int
    image: Optional[str]
    restart: Optional[str]
    volumes: Optional[List[str]]
    ports: Optional[List[str]]
    expose: Optional[List[int]]
    command: Optional[List[str]]
    environment: Optional[Dict[str, Union[str, int]]]

    @staticmethod
    def from_kubernetes(yaml: Dict[str, Any]) -> Tuple['DockerComposeService', Dict[str, None]]:
        return DockerComposeService(
            name=yaml['metadata']["name"],
            kubernetes_type=yaml['kind'],
            image=yaml["spec"]["template"]["spec"]["containers"][0]["image"],
            restart=MAP_CHOICES_RESTART[yaml["spec"]["template"]["spec"].get("restartPolicy", "Always")],
            volumes=[f"{volume['name']}:{volume['mountPath']}" for volume in yaml["spec"]["template"]["spec"]["containers"][0]["volumeMounts"]]
            if yaml["spec"]["template"]["spec"]["containers"][0].get("volumeMounts") is not None else [],
            environment={
                env['name']: env['value'] for env in yaml["spec"]["template"]["spec"]["containers"][0]["env"]
            } if yaml["spec"]["template"]["spec"]["containers"][0].get("env") is not None
            else {},
            command=yaml["spec"]["template"]["spec"]["containers"][0].get("command", []) + yaml["spec"]["template"]["spec"]["containers"][0].get("args", []),
            replicas=yaml["spec"].get("replicas", 1),
            expose=[
                r["containerPort"] for r in yaml["spec"]["template"]["spec"]["containers"][0].get("ports", [])
            ] if not yaml['metadata']["name"].endswith('balancer') else [],
            ports=[
                f'{r["containerPort"]}:{r["containerPort"]}' for r in yaml["spec"]["template"]["spec"]["containers"][0].get("ports", [])
            ] if yaml['metadata']["name"].endswith('balancer') else [],

        ), {volume['name']: None for volume in yaml["spec"]["template"]["spec"]["containers"][0]["volumeMounts"]} if yaml["spec"]["template"]["spec"]["containers"][0].get("volumeMounts") is not None else {}

    def to_json(self, common_dependencies: List[str]) -> Dict[str, Any]:
        return {
            self.name: {
                "image": self.image,
                "restart": self.restart,
                "volumes": self.volumes,
                "deploy": {
                    "replicas": self.replicas,
                },
                "ports": self.ports,
                "expose": self.expose,
                "command": self.command,
                "environment": self.environment,
                "depends_on": common_dependencies,
            },
        } if self.replicas != 0 else {}


class DockerComposeFile(BaseModel):
    volumes: Dict[str, None]
    common_dependencies: List[str]
    version: str
    services: List[DockerComposeService]

    @staticmethod
    def from_kubernetes(file: Iterable[Dict[str, Any]]) -> 'DockerComposeFile':
        volumes_list: Dict[str, None] = {}
        services_list: List[DockerComposeService] = []
        common_dependencies: List[str] = []
        for resource in file:
            if resource['kind'] not in KUBERNETES_OBJECTS__TRANSFORM_TO_DCF__TYPE:
                continue
            service, volumes = DockerComposeService.from_kubernetes(resource)
            volumes_list.update(**volumes)
            services_list.append(service)
            if resource['kind'] in KUBERNETES_OBJECTS__DEPENDS_ON_OBJECT__TYPE:
                # No need in worker or api if no DB or broker exists
                common_dependencies.append(service.name)

        return DockerComposeFile(
            version="3.8",
            volumes=volumes_list,
            services=services_list,
            common_dependencies=common_dependencies,
        )

    def to_json(self) -> Dict[str, Any]:
        services_dict: Dict[str, Any] = {}
        for service in self.services:
            service_json = service.to_json(
                common_dependencies=self.common_dependencies if service.kubernetes_type not in KUBERNETES_OBJECTS__DEPENDS_ON_OBJECT__TYPE else [],
            )
            services_dict.update(**service_json)
        return {
            "version": "3.8",
            "volumes": self.volumes,
            "services": services_dict,
        }



import contextlib
import io
import os
import shutil
import subprocess
import sys
import traceback
from typing import Generator, List, Optional, Callable, Union, Dict, Set, NamedTuple

from ul_py_tool.commands.conf import MAX_LINE_LEN
from ul_py_tool.utils.arg_files_glob import arg_file_info
from ul_py_tool.utils.colors import FG_RED, NC, FG_GREEN, FG_GRAY, FG_BLUE, FG_YELLOW
from ul_py_tool.utils.write_stdout import write_stdout

TChk = Optional[Callable[['subprocess.CompletedProcess[bytes]'], None]]


class StepError(Exception):
    pass


class StepFileErr(NamedTuple):
    err: str
    col: Optional[int]
    code: Optional[str]


class Step:
    def __init__(self, name: str, n: int) -> None:
        self._n = n
        self._name = name
        self._raised = False
        self._errors_in_file: Dict[str, Dict[int, Set[StepFileErr]]] = dict()
        self._skipped = False

        write_stdout('')
        self._print(f'{FG_BLUE}{self._name}{NC}')

        self._errors: Set[str] = set()
        self._warns: Set[str] = set()

    def _print(self, txt: str) -> None:
        write_stdout(f'{FG_BLUE}[{self._n:0>2}]{NC} :: {txt}')

    def skip(self) -> None:
        if self._skipped:
            return
        self._skipped = True
        self._print(f'{FG_GRAY}SKIPPED{NC}')

    def add_error(self, message: str) -> None:
        self._errors.add(message)

    def add_warn(self, message: str) -> None:
        self._warns.add(message)

    def _internal_use_raise_if_has_errors(self) -> None:
        assert not self._raised
        self._raised = True

        has_errors = False
        max_len_of_file = 0
        errors_list = []
        for file, error_lines in self._errors_in_file.items():
            for line, errors in error_lines.items():
                for err in errors:
                    file_info = arg_file_info(file, line=line, col=err.col)
                    file_err = f'{err.err} {f"{FG_GRAY}({err.code}){NC}" if err.code is not None else ""}'
                    max_len_of_file = max(max_len_of_file, len(file_info))
                    errors_list.append((file_info, file_err))

        for file_info, file_err in errors_list:
            write_stdout(f'{file_info: <{max_len_of_file}} :: {file_err}')

        has_errors = len(errors_list) > 0 or has_errors
        has_warns = len(self._warns) > 0

        for err_str in self._errors:
            has_errors = True
            write_stdout(err_str)

        for warn_str in self._warns:
            write_stdout(f'{FG_YELLOW}{warn_str}{NC}')

        if has_errors:
            raise StepError()

        if not self._skipped:
            self._print(f'{FG_GREEN if not has_warns else FG_YELLOW}DONE{NC}')

    def add_error_in_file(self, file: str, err: str, *, line: int = 0, col: Optional[int] = None, code: Optional[str] = None) -> None:
        if file not in self._errors_in_file:
            self._errors_in_file[file] = dict()
        if line not in self._errors_in_file[file]:
            self._errors_in_file[file][line] = set()
        self._errors_in_file[file][line].add(StepFileErr(col=col, code=code, err=err))

    def run_cmd(self, cmd: Union[List[str], str], chk: TChk = None, ignore_error: bool = False, print_std: bool = True) -> 'subprocess.CompletedProcess[bytes]':
        ret_code = -1

        kwargs = dict()
        if chk is None and print_std:
            kwargs['stderr'] = sys.stderr
            kwargs['stdout'] = sys.stdout
        cmd_log = cmd if isinstance(cmd, str) else (" ".join(cmd))
        write_stdout(f'run cmd :: {FG_GRAY}{cmd_log if len(cmd_log) < MAX_LINE_LEN else f"{cmd_log[:MAX_LINE_LEN]}..."}{NC}')

        try:
            subprocess_result = subprocess.run(  # type: ignore
                cmd,
                shell=isinstance(cmd, str),
                check=False,
                capture_output=chk is not None or not print_std,
                cwd=os.getcwd(),
                env={
                    **os.environ,
                    "COLOR": '1',
                    "COLORS": '1',
                    "FORCE_COLOR": '1',
                    "FORCE_COLORS": '1',
                },
                **kwargs,
            )

            ret_code = subprocess_result.returncode
            if chk is not None:
                if print_std:
                    shutil.copyfileobj(io.StringIO(subprocess_result.stdout.decode()), sys.stdout)
                    shutil.copyfileobj(io.StringIO(subprocess_result.stderr.decode()), sys.stderr)
                chk(subprocess_result)
            else:
                subprocess_result.check_returncode()
        except Exception as e:  # noqa: B902
            if not isinstance(e, subprocess.CalledProcessError):
                write_stdout(f'{FG_RED}ERROR:{NC} {type(e).__name__} :: {e}')
                write_stdout(traceback.format_exc())
            if not ignore_error:
                write_stdout(f'run cmd :: {FG_RED}ERROR{NC} exit code {ret_code}')
                raise StepError()
            write_stdout(f'run cmd :: {FG_YELLOW}DONE with ignored error{NC}')
            return subprocess_result
        write_stdout(f'run cmd :: {FG_GREEN}DONE{NC}')

        return subprocess_result


class Stepper:
    def __init__(self) -> None:
        self._index = 0

    @contextlib.contextmanager
    def step(self, name: str, print_error: bool = True, ignore_error: bool = False) -> Generator[Step, None, None]:
        self._index += 1
        stp = Step(name, self._index)
        try:
            yield stp
            stp._internal_use_raise_if_has_errors()  # noqa
        except StepError:
            raise
        except Exception as e:  # noqa: B902
            if print_error:
                write_stdout(f'{FG_RED}ERROR:{NC} {type(e).__name__} :: {e}')
            if not ignore_error:
                raise
            write_stdout(f'{FG_GREEN}>> {name} :: DONE {NC}{FG_YELLOW}with ignored error{NC}')



import argparse
import os
from typing import Union

import yaml

from ul_py_tool.utils.aseembly import AssemblyFile


def arg_str2yaml(value: Union[str, int, bool]) -> AssemblyFile:
    if not isinstance(value, str):
        raise argparse.ArgumentTypeError('invalid type')
    file_path = value
    if not value.startswith('/'):
        file_path = os.path.join(os.getcwd(), value)
    if not os.path.isfile(file_path):
        raise argparse.ArgumentTypeError(f'File {file_path} not found')
    with open(file_path, 'r') as f:
        assembly = yaml.safe_load(f)

    return AssemblyFile(**assembly)



from os import path
from setuptools import setup, find_packages

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='ul-py-tool',
    version='1.15.49',
    description='Python ul py tool',
    author='Unic-lab',
    author_email='',
    url='https://gitlab.neroelectronics.by/unic-lab/libraries/common-python-utils/ul-py-tool.git',
    packages=find_packages(include=['ul_py_tool*']),
    platforms='any',
    package_data={
        '': [
            'conf/*',
        ],
        'ul_py_tool': [
            'py.typed',
        ],
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            'ulpytool=ul_py_tool.main:main',
        ],
    },
    include_package_data=True,
    install_requires=[
        "numpy==1.23.3",
        "pandas==1.4.3",
        "pydantic[mypy]==1.10.2",
        "PyYAML==6.0",
        "colored==1.4.3",
        "rich==12.6.0",
        "tomli==2.0.1",
        "requests==2.28.1",

        "deepdiff==5.8.1",

        "mypy==0.982",
        "types-pyyaml==6.0.11",
        "types-pytz==2022.1.2",
        "types-python-dateutil==2.8.19",
        "types-setuptools==63.4.0",
        "typing-extensions==4.3.0",
        "data-science-types==0.2.23",
        "types-requests==2.28.8",

        "pycodestyle==2.9.1",
        "flake8==5.0.4",
        "flake8-commas==2.1.0",
        "flake8-noqa==1.2.8",
        "flake8-polyfill==1.0.2",
        "flake8-blind-except==0.2.1",
        "flake8-tidy-imports==4.8.0",
        "flake8-self==0.2.2",
        "flake8-super-call==1.0.0",
        "flake8-bugbear==22.7.1",
        "flake8-print==5.0.0",
        "pep8-naming==0.13.1",  # flask plugin
        "isort[colors]==5.10.1",
        "black==22.10.0",
        "yamllint==1.28.0",
        "ruff==0.0.198",

        "pytest==7.1.3",
        "pytest-cov==3.0.0",
        "python-gitlab==3.8.0",
        "kubernetes==23.6.0",

        "wheel==0.37.1",
        "twine==4.0.1",
        "setuptools==63.4.2",
    ],
)



import re
from dataclasses import dataclass
from typing import Optional, Union, Tuple

VERSION_RE = re.compile(r'^(==|>|<|<=|>=|~=|\^=)?\s*(\d+)\.(\d+)(?:\.(\d+))?(.*?)$')


@dataclass
class SemVer:
    major: int
    minor: int
    patch: Optional[int] = None
    variant: str = ''

    def __lt__(self, other: Union[str, 'SemVer']) -> bool:
        other_ver = SemVer.from_string(other) if isinstance(other, str) else other
        if other_ver is None:
            raise ValueError(f'invalid value {other}')
        if self.major > other_ver.major:
            return False
        if self.major < other_ver.major:
            return True
        if self.minor > other_ver.minor:
            return False
        if self.minor < other_ver.minor:
            return True
        if self.patch is None:
            if other_ver.patch is not None:
                return True
        else:
            if other_ver.patch is None:
                return False
        if other_ver.patch is not None and self.patch is not None:
            if self.patch > other_ver.patch:
                return False
            if self.patch < other_ver.patch:
                return True
        if self.variant and not other_ver.variant:
            return False
        if other_ver.variant and not self.variant:
            return True
        return False

    def __gt__(self, other: Union[str, 'SemVer']) -> bool:
        other_ver = SemVer.from_string(other) if isinstance(other, str) else other
        if other_ver is None:
            raise ValueError(f'invalid value {other}')
        return other_ver < self

    def __le__(self, other: Union[str, 'SemVer']) -> bool:
        other_ver = SemVer.from_string(other) if isinstance(other, str) else other
        if other_ver is None:
            raise ValueError(f'invalid value {other}')
        return self == other or self < other

    def __ge__(self, other: Union[str, 'SemVer']) -> bool:
        other_ver = SemVer.from_string(other) if isinstance(other, str) else other
        if other_ver is None:
            raise ValueError(f'invalid value {other}')
        return other_ver == self or other_ver < self

    @staticmethod
    def from_string(version_string: str) -> Optional['SemVer']:
        version_string = version_string.strip()
        m = VERSION_RE.match(version_string)
        if m is None:
            return None
        _1, ma, mi, pa, va = m.groups()
        return SemVer(
            major=int(ma),
            minor=int(mi),
            patch=int(pa) if pa is not None else None,
            variant=va,
        )

    @staticmethod
    def from_string_wop(version_string: str) -> Optional[Tuple[str, 'SemVer']]:
        version_string = version_string.strip()
        m = VERSION_RE.match(version_string)
        if m is None:
            return None
        op, ma, mi, pa, va = m.groups()
        return op, SemVer(
            major=int(ma),
            minor=int(mi),
            patch=int(pa) if pa is not None else None,
            variant=va,
        )

    def __str__(self) -> str:
        return f'{self.major}.{self.minor}{f".{self.patch}" if self.patch is not None else ""}{self.variant}'


def test_semver() -> None:
    assert SemVer.from_string('1.3.2') == SemVer(1, 3, 2)
    assert SemVer.from_string('1.3') == SemVer(1, 3)
    assert SemVer.from_string('1.3-some') == SemVer(1, 3, variant='-some')
    assert str(SemVer(1, 3, 2)) == '1.3.2'
    assert str(SemVer(1, 3)) == '1.3'
    assert str(SemVer(1, 3)) != '1.4'
    assert str(SemVer(1, 3, variant='-some')) == '1.3-some'

    assert SemVer(1, 3, 2) > SemVer(1, 3, 1)
    assert SemVer(1, 3, 2) > SemVer(1, 3, 0)
    assert SemVer(1, 3, 2) > SemVer(1, 2, 0)
    assert SemVer(1, 3, 2) > SemVer(1, 2, 10)
    assert SemVer(1, 3, 2) > SemVer(0, 10, 10)
    assert SemVer(1, 3, 2) > SemVer(0, 3, 3)
    assert SemVer(1, 3, 2) > SemVer(0, 3)
    assert SemVer(1, 3, 2) > SemVer(1, 3)

    assert SemVer(1, 3, 2) >= SemVer(1, 3, 1)
    assert SemVer(1, 3, 2) >= SemVer(1, 3, 0)
    assert SemVer(1, 3, 2) >= SemVer(1, 2, 0)
    assert SemVer(1, 3, 2) >= SemVer(1, 2, 10)
    assert SemVer(1, 3, 2) >= SemVer(0, 10, 10)
    assert SemVer(1, 3, 2) >= SemVer(0, 3, 3)
    assert SemVer(1, 3, 2) >= SemVer(0, 3)
    assert SemVer(1, 3, 2) >= SemVer(1, 3)
    assert SemVer(1, 3, 2) >= SemVer(1, 3, 2)

    assert SemVer(1, 3, 1) < SemVer(1, 3, 2)
    assert SemVer(1, 3, 0) < SemVer(1, 3, 2)
    assert SemVer(1, 2, 0) < SemVer(1, 3, 2)
    assert SemVer(1, 2, 10) < SemVer(1, 3, 2)
    assert SemVer(0, 10, 10) < SemVer(1, 3, 2)
    assert SemVer(0, 3, 3) < SemVer(1, 3, 2)
    assert SemVer(0, 3) < SemVer(1, 3, 2)
    assert SemVer(1, 3) < SemVer(1, 3, 2)

    assert SemVer(1, 3, 1) < '1.3.2'
    assert SemVer(1, 3, 0) < '1.3.2'
    assert SemVer(1, 2, 0) < '1.3.2'
    assert SemVer(1, 2, 10) < '1.3.2'
    assert SemVer(0, 10, 10) < '1.3.2'
    assert SemVer(0, 3, 3) < '1.3.2'
    assert SemVer(0, 3) < '1.3.2'
    assert SemVer(1, 3) < '1.3.2'
    assert not (SemVer(1, 3) < SemVer(1, 3))

    assert SemVer(1, 3, 1) <= '1.3.2'
    assert SemVer(1, 3, 0) <= '1.3.2'
    assert SemVer(1, 2, 0) <= '1.3.2'
    assert SemVer(1, 2, 10) <= '1.3.2'
    assert SemVer(0, 10, 10) <= '1.3.2'
    assert SemVer(0, 3, 3) <= '1.3.2'
    assert SemVer(0, 3) <= '1.3.2'
    assert SemVer(1, 3) <= '1.3.2'

    assert SemVer(1, 3, 1) == SemVer(1, 3, 1)
    assert SemVer(1, 3, 0) == SemVer(1, 3, 0)
    assert SemVer(1, 2, 0) == SemVer(1, 2, 0)
    assert SemVer(1, 2, 10) == SemVer(1, 2, 10)
    assert SemVer(0, 10, 10) == SemVer(0, 10, 10)
    assert SemVer(0, 3, 3) == SemVer(0, 3, 3)



import argparse
import os

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.commands.conf import THIS_LIB_PATH, MYPY_CONFIG, PYENV_PY_VERSION, PRE_PUSH_DEST, PRE_COMMIT_DEST
from ul_py_tool.utils.arg_str2bool import arg_str2bool
from ul_py_tool.utils.fs import copy_if_destination_is_absent
from ul_py_tool.utils.step import Stepper


class CmdInstall(Cmd):
    force_replace: bool = False

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--force', dest='force_replace', type=arg_str2bool, required=False, default=False)

    def run(self) -> None:
        stepper = Stepper()

        with stepper.step('fix rights of git-hooks') as stp:
            stp.run_cmd(['chmod', '755', '-R', os.path.join(os.getcwd(), '.git', 'hooks')], ignore_error=True)

        with stepper.step('init git pre-push hook'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', 'git.hook.pre-push.sh'), PRE_PUSH_DEST, replace=True)

        with stepper.step('init git pre-commit hook'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', 'git.hook.pre-commit.sh'), PRE_COMMIT_DEST, replace=True)

        with stepper.step('init .gitignore'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', 'gitignore'), os.path.join(os.getcwd(), '.gitignore'), replace=self.force_replace)

        with stepper.step('init .editorconfig'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', 'editorconfig'), os.path.join(os.getcwd(), '.editorconfig'), replace=self.force_replace)

        with stepper.step('init .gitattributes'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', 'gitattributes'), os.path.join(os.getcwd(), '.gitattributes'), replace=self.force_replace)

        with stepper.step('init python-version'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', 'python-version'), os.path.join(os.getcwd(), PYENV_PY_VERSION), replace=self.force_replace)

        with stepper.step('init mypy'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', MYPY_CONFIG), os.path.join(os.getcwd(), MYPY_CONFIG), replace=self.force_replace)

        with stepper.step('init pytest'):
            copy_if_destination_is_absent(os.path.join(THIS_LIB_PATH, 'conf', 'pytest.ini'), os.path.join(os.getcwd(), 'pytest.ini'), replace=self.force_replace)

        with stepper.step('fix rights of git-hooks') as stp:
            stp.run_cmd(['chmod', '755', '-R', os.path.join(os.getcwd(), '.git', 'hooks')], ignore_error=True)

        # with stepper.step('MYPY') as stp:
        #     stp.run_cmd([
        #         'mypy',
        #         '--install-types',
        #         '--non-interactive',
        #     ], ignore_error=True)
